

#ifndef SY_H
#define SY_H

#include "defs.h"

#define LAPIS_VERSION "0.8"

void init_lapis ();
bool process_lapis (int fd, char *localpath);

enum {
        STACKITEM_NULL=0,
        STACKITEM_REAL=1,
        STACKITEM_INTEGER=2,
        STACKITEM_STRING=3,
        STACKITEM_SYMBOL=4,
        STACKITEM_CODE=5,
        STACKITEM_STYLE=6,
        STACKITEM_IMAGE=7,
        STACKITEM_ARRAY=8,
        STACKITEM_DICTIONARY=9,
        STACKITEM_FONT=10,
        STACKITEM_MATRIX=11,
        STACKITEM_BOOLEAN=12,
        STACKITEM_ARY_BOUNDARY=13,
        STACKITEM_CLASS=14,
        STACKITEM_PARAMSTART=15,
        STACKITEM_PARAMEND=16,
        STACKITEM_TABLE=17,
};

#define ATTR_BOLD 1
#define ATTR_ITALIC 2
#define ATTR_UNDERLINE 4
#define ATTR_SUPERSCRIPT 8
#define ATTR_SUBSCRIPT 16
#define ATTR_STRIKEOUT 32
#define ATTR_TITLE 64
#define ATTR_PLAIN 128

typedef unsigned char uchar;

typedef struct font {
	char *name;
	int num;
	char width, height;
	unsigned int consolefont : 1;
	unsigned int bold : 1;
	unsigned int italic : 1;
	unsigned int fixed : 1;
	uchar *data;
	// struct font *next;
	short refcount;
} Font;

typedef struct img {
	short width, height, x, y;
	uchar *pixels;
	unsigned long color;
	Font *font;

	short refcount;
} Image;

typedef struct array {
	unsigned int	size;
	unsigned int	allocated_size;
	void	**values;
	short 	refcount;
} Array;

typedef struct matrix {
	short width, height;
	long double *values;
	short refcount;
} Matrix;

typedef struct dictionaryitem {
        char    *name;
        void	*value;
        struct dictionaryitem *next;
} DictionaryItem;

typedef struct dictionary {
	DictionaryItem *list;
	short refcount;
} Dictionary;

typedef struct ti {
	char	type;	// 0 = string, 1 = number
	union	{
		char 	*string;
		long double	number;
	};
	struct ti *next;
} TableItem;

typedef struct tr {
	TableItem*items;
	TableItem *last;
	struct tr *next;
	int total_items;
} TableRow;

typedef struct tt {
	TableRow	*header;
	TableRow	*rows;
	TableRow	*last;
#define MAX_COLUMNS 256
	short 	column_widths [MAX_COLUMNS];
	bool 	column_is_numeric [MAX_COLUMNS];
	short	total_columns;
	short	refcount;
	int 	total_rows;
} Table;

typedef struct stackim {
        char    type;

	union {
		int     integer;
		long double  real;	// 8 bytes
		uchar*   string;
		Image	*image;
		Font	*font;	
		Dictionary	*dictionary;
		Array	*array;
		Matrix	*matrix;
		Table	*table;
	} datum;

        unsigned int bold : 1;
	unsigned int italic : 1;
	unsigned int underline : 1;
	unsigned int super : 1;
	unsigned int sub : 1;
	unsigned int strikeout : 1;
	unsigned int title : 1;
	int size : 4; // relative change e.g. big= +1

        struct stackim  *next;
}
StackItem;

typedef int (*non_method_func) (short cmdnum, StackItem **, FILE *);
typedef int (*method_func) (short cmdnum, StackItem *self, StackItem *in, StackItem **out, FILE *);

struct cmdinfo {
	char *name;
	short cmdnum;
	short num_params; // incoming
	non_method_func func;
	char *to_print;
	char *print_with_style; // e.g. <p style="foo">
	unsigned char flags;
	char sizechange;
};

typedef struct method {
	char *name;
	method_func func;
	short nparams;
	short cmdnum;
	struct method *next;
} Method;

typedef struct class {
	char *name;
	Method *methods[27];
	struct class *next;
} Class;

enum {
	LAPIS_SUCCESS=0,
	LAPIS_ERR_NONE=0,
	LAPIS_ERR_UNDERFLOW=1,
	LAPIS_ERR_DIVIDE=2,
	LAPIS_ERR_TYPE=3,
	LAPIS_ERR_UNKNOWN=4,
	LAPIS_ERR_NOMEM=5,
	LAPIS_ERR_NOFONT=6,
	LAPIS_ERR_ARRAYBOUND=7,
	LAPIS_ERR_NOFUNC=8,
	LAPIS_ERR_ARRAYINDEX=9,
	LAPIS_ERR_INTERNAL=10,
	LAPIS_ERR_RESERVED=11,
	LAPIS_ERR_UNIMPLEMENTED=12,
	LAPIS_ERR_UNKNOWN_METHOD=13,
	LAPIS_ERR_METHOD_WITH_DOT=14,
	LAPIS_ERR_CANNOT_GET_CLASS=15,
	LAPIS_ERR_FATAL=16,
	LAPIS_ERR_PARAMBOUND=17,
	LAPIS_ERR_NO_CONSTRUCTOR=18,
	LAPIS_ERR_NEED_PARAMS=19,
	LAPIS_ERR_NO_OBJ=20,
	LAPIS_OK_REMOVE_SELF=21,
	LAPIS_ERR_DATA_RECURSION=22,
	LAPIS_ERR_MATRIXROW=23,
	LAPIS_ERR_MATRIXCOLUMN=24,
	LAPIS_ERR_BADREAL=25,
	LAPIS_ERR_LOAD=26,
	LAPIS_OK_EXIT=27,
	LAPIS_ERR_BADVARMETHOD=28,
	LAPIS_ERR_VARIABLE=29
};




enum {
	CMD_LBRACKET=1,
	CMD_RBRACKET=2,
	CMD_NOT=3,
	CMD_NOTEQUAL=4,
	CMD_REMAINDER=5,
	CMD_AND=6,
	CMD_BOOLAND=7,
	CMD_MULT=8,
	CMD_ADD=9,
	CMD_INCREMENT=10,
	CMD_SUB=11,
	CMD_DECREMENT=12,
	CMD_DIV=13,
	CMD_2DUP=14,
	CMD_2DROP=15,
	CMD_LT=16,
	CMD_LSHIFT=17,
	CMD_LE=18,
	CMD_EQUAL=19,
	CMD_GT=20,
	CMD_GE=21,
	CMD_RSHIFT=22,
	CMD_XOR=23,
	CMD_BOOLXOR=24,
	CMD_OR=25,	
	CMD_BOOLOR=26,
	CMD_ACOS=27,
	CMD_ARRAY=28,
	CMD_ASIN=29,
	CMD_ATAN=30,
	CMD_BIGBEGIN=31,
	CMD_BIGEND=32,
	CMD_BIG=33,
	CMD_BODY=34,
	CMD_BOLD=35,
	CMD_BOLDBEGIN=36,
	CMD_BOLDEND=37,
	CMD_BR=38,
	CMD_CBRT=39,
	CMD_CENTERBEGIN=40,
	CMD_CENTEREND=41,
	CMD_CONTENT_HTML=42,
	CMD_CONTENT_JPEG=43,
	CMD_COS=44,
	CMD_CONSTANT_E=45,
	CMD_CONSTANT_SQRT2=46,
	CMD_DEF=47,
	CMD_DESCRIPTIONSBEGIN=48,
	CMD_DESCRIPTIONSCONTENTS=49,
	CMD_DESCRIPTIONSEND=50,
	CMD_DESCRIPTIONSTITLE=51,
	CMD_DIVBEGIN=52,
	CMD_DIVEND=53,
	CMD_DROP=54,
	CMD_DUP=55,
	CMD_EXCH=56,
	CMD_FALSE=57,
	CMD_FILLRECT=58,
	CMD_FONTBEGIN=59,
	CMD_FONTEND=60,
	CMD_FONTFACEBEGIN=61,
	CMD_FONTFACEEND=62,
	CMD_FONTSIZEBEGIN=63,
	CMD_FONTSIZEEND=64,
	CMD_GET=65,
	CMD_GETCOLOR=66,
	CMD_GETENV=67,
	CMD_GETCONSOLEFONT=68,
	CMD_HRULE=69,
	CMD_H1BEGIN=70,
	CMD_H1END=71,
	CMD_H2BEGIN=72,
	CMD_H2END=73,
	CMD_H3BEGIN=74,
	CMD_H3END=75,
	CMD_H4BEGIN=76,
	CMD_H4END=77,
	CMD_H5BEGIN=78,
	CMD_H5END=79,
	CMD_H6BEGIN=80,
	CMD_H6END=81,
	CMD_IMAGE=82,
	CMD_INDENTBEGIN=83,
	CMD_INDENTEND=84,
	CMD_ITALIC=85,
	CMD_ITALICBEGIN=86,
	CMD_ITALICEND=87,
	CMD_ITEM=88,
	CMD_LINETO=89,
	CMD_LINK=90,
	CMD_LISTBEGIN=91,
	CMD_LISTEND=92,
	CMD_LN=93,
	CMD_DICTIONARY=94,
	CMD_MATRIX=95,
	CMD_MOVETO=96,
	CMD_NEG=97,
	CMD_NIP=98,
	CMD_NOSTYLE=99,
	CMD_NULL=100,
	CMD_NUMBEREDBEGIN=101,
	CMD_NUMBEREDEND=102,
	CMD_OVER=103,
	CMD_PAR=104,
	CMD_PARABEGIN=105,
	CMD_PARAEND=106,
	CMD_CONSTANT_PI=107,
	CMD_POW=108,
	CMD_PRINT=109,
	CMD_PUT=110,
	CMD_RIGHTBEGIN=111,
	CMD_RIGHTEND=112,
	CMD_ROT=113,
	CMD_SETCOOKIE=114,
	CMD_SETCOLOR=115,
	CMD_SETFONT=116,
	CMD_SHOW=117,
	CMD_SIN=118,
	CMD_SMALLBEGIN=119,
	CMD_SMALLEND=120,
	CMD_SMALL=121,
	CMD_SPANBEGIN=122,
	CMD_SPANEND=123,
	CMD_SQRT=124,
	CMD_STRIKEBEGIN=125,
	CMD_STRIKEEND=126,
	CMD_STRIKEOUT=127,
	CMD_STRINGCHAR=128,
	CMD_STRINGLEFT=129,
	CMD_STRINGRIGHT=130,
	CMD_STYLE=131,
	CMD_SUBSCRIPT=132,
	CMD_SUBSTR=133,
	CMD_SUPERSCRIPT=134,
	CMD_SWAP=135,
	CMD_TABLEBEGIN=136,
	CMD_TABLECOLUMN=137,
	CMD_TABLEROW=138,
	CMD_TABLEEND=139,
	CMD_TABLEHDR=140,
	CMD_TAN=141,
	CMD_TITLE=142,
	CMD_TRUE=143,
	CMD_UNDERLINE=144,
	CMD_UNDERLINEBEGIN=145,
	CMD_UNDERLINEEND=146,
	CMD_VERSION=147,
	CMD_WRITE=148,
	CMD_WRITELN=149,
	CMD_LPARENS=150,
	CMD_RPARENS=151,
	CMD_LBRACE=152,
	CMD_RBRACE=153,
	CMD_STRINGTRUNC=154,
	CMD_INVERT,
};
#endif

